unit main;

{$assertions on}

{$mode objfpc}{$H+}
//{$mode delphi}
interface

uses
  Classes, SysUtils, types, FileUtil, Forms, Controls, Graphics, Dialogs,
  Buttons, StdCtrls, LCLType, LCLProc, ExtCtrls, LCLIntf,
  ComCtrls, LazFileUtils, SynEdit;

type
  { TForm1 }

  TTHEdit = class;
  TTextHintOverlay= class;

  TForm1 = class(TForm)
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    Button4: TButton;
    Button5: TButton;
    Edit1: TEdit;
    Memo1: TMemo;
    Panel1: TPanel;
    Panel2: TPanel;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure Button5Click(Sender: TObject);
    procedure Edit1Enter(Sender: TObject);
    procedure Edit1Exit(Sender: TObject);
    procedure Edit1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);

  private
    { private declarations }
    THEdit: TTHEdit;
    THO: TTextHintOverlay;
  public
    { public declarations }
    //property Foo: Integer read GetFoo write SetFoo;
  end;


  { TTextHintOverlay }
  TCustomEdit = class(StdCtrls.TCustomEdit)
  end;


  {$ifdef LCLGTK2}{$define thdebug}{$endif}

  TTextHintOverlay = class(TCustomControl)
  private
    FEdit: TCustomEdit;
    function CanShow: Boolean;
    procedure SetParentToNil({%H-}Dummy: PtrInt);
    procedure QueueSetParentToNil;
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
                        X, Y: Integer); override;
    procedure DoEnter; override;
    procedure Paint; override;
    procedure RealSetText(const AValue: TCaption); override;
  public
    constructor Create(AOwner: TComponent; AEdit: TCustomEdit); overload;
    destructor Destroy; override;
    //procedure Show; override;
    //procedure Hide; override;
    procedure ShowOverlay;
    procedure HideOverlay;
    procedure SetFocus; override;

  end;

  { TTHEdit }

  TTHEdit = class(TCustomEdit)
  private
    FTextHint: String;
    TextHintOverlay: TTextHintOverlay;
    function CreateTextHintOverlay(ACaption: TCaption): TTextHintOverlay;
    procedure SetTextHint(AValue: String);
    procedure ShowTextHint;
    procedure HideTextHint;
    function TextHintVisible: Boolean;
  protected
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure DoSetBounds(ALeft, ATop, AWidth, AHeight: integer); override;
    procedure Loaded; override;
    procedure TextChanged; override;
    procedure SetParent(NewParent: TWinControl); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property TextHint: String read FTextHint write SetTextHint;
  end;


var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }

const
  StrangeFn = 'C:\Users\Bart\LazarusProjecten\bugs\FileUtf8\' +
              #$E0#$B8#$9A#$E0#$B8#$A5#$E0#$B8#$B9#$E0#$B9#$80#$E0#$B8#$9A#$E0#$B8#$AD + '.txt';

{ TTextHintOverlay }

function TTextHintOverlay.CanShow: Boolean;
begin
  Result := Assigned(FEdit)
            and Assigned(FEdit.Parent)
            and FEdit.Visible
            and FEdit.Enabled
            and (Caption <> '')
            and (([csDesigning,csLoading] * FEdit.ComponentState) = []);
  //writeln('TTextHintOverlay.CanShow: Result = ',Result);
end;

procedure TTextHintOverlay.SetParentToNil(Dummy: PtrInt);
begin
  Parent := nil;
end;

procedure TTextHintOverlay.QueueSetParentToNil;
begin
  Application.QueueAsyncCall(@SetParentToNil, 0);
end;

procedure TTextHintOverlay.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  //writeln('TTextHintOverlay.MouseDown');
  //HideOverlay;
  FEdit.SetFocus;
  FEdit.MouseDown(Button, Shift, X, Y);
end;

procedure TTextHintOverlay.DoEnter;
begin
  inherited DoEnter;
  HideOverlay;
  if Assigned(FEdit) then FEdit.SetFocus;
end;


procedure TTextHintOverlay.Paint;
var
  PaintRect: TRect;
  Style: TTextStyle;
begin
  //writeln('TTextHintOverlay.Paint: FEdit = ',DbgSName(FEdit));
  if not CanShow then Exit;
  inherited Paint;
  PaintRect := ClientRect;
  //Canvas.Brush.Color := FEdit.Color;
  //Canvas.Pen.Color := clRed;
  //Canvas.FrameRect(PaintRect);
  //InflateRect(PaintRect, FEdit.BorderSpacing.InnerBorder, FEdit.BorderSpacing.InnerBorder);
  Style.Alignment := FEdit.Alignment;
  Style.SingleLine := True;
  Style.Layout := tlCenter;
  Style.EndEllipsis := True;
  Style.RightToLeft := (BiDiMode in [bdRightToLeft, bdRightToLeftReadingOnly]);
  Style.Clipping := False;
  Style.ShowPrefix := False;
  Style.ExpandTabs := True;
  Style.Wordbreak := False;
  Style.SystemFont := False;
  Style.Opaque := True;
  Canvas.TextRect(PaintRect, 2, 0, Caption, Style);
end;



procedure TTextHintOverlay.RealSetText(const AValue: TCaption);
var
  OldValue: TCaption;
begin
  OldValue := Text;
  inherited RealSetText(AValue);
  if (OldValue <> AValue) then
  begin
    if (AValue <> '') then
      Invalidate
    else
      HideOverlay;
  end;
end;


procedure TTextHintOverlay.SetFocus;
begin
  if Assigned(FEdit) then FEdit.SetFocus;
  if Visible then HideOverlay;
end;

constructor TTextHintOverlay.Create(AOwner: TComponent; AEdit: TCustomEdit);
begin
  inherited Create(AOwner);
  FEdit := AEdit;
  ControlStyle := ControlStyle + [csNoDesignVisible, csNoFocus];
  TabStop := False;
  ParentColor := False;
end;

destructor TTextHintOverlay.Destroy;
begin
  inherited Destroy;
end;

procedure TTextHintOverlay.ShowOverlay;
var
  BRect: TRect;
begin
  if not CanShow then Exit;
  BRect := FEdit.BoundsRect;
  //if FEdit.BorderStyle = bsSingle then InflateRect(BRect, -2, -2);
  BorderStyle := FEdit.BorderStyle;
  BoundsRect := BRect;
  {$ifdef thdebug}
  Top := Top + ((1 * FEdit.Height) div 3);
  //Left := Left + (FEdit.Width div 2);
  {$endif}
  Font.Assign(FEdit.Font);
  Font.Style := [fsItalic];
  Font.Color := clGrayText;
  BidiMode := FEdit.BiDiMode;
  Cursor := FEdit.Cursor;
  Color := FEdit.GetRGBColorResolvingParent;
  Parent := FEdit.Parent;
  //Enabled := False;
  Visible := True;
  BringToFront;
end;

procedure TTextHintOverlay.HideOverlay;
begin
  //writeln('TTextHintOverlay.HideOverlay: Visible = ',Visible);
  if not Visible then Exit;
  Visible := False;
  //GTK2 barfs if you do Parent := nil, when this procedure is called from within MouseDown...
  QueueSetParentToNil;
end;

{ TTHEdit }

function TTHEdit.CreateTextHintOverlay(ACaption: TCaption): TTextHintOverlay;
begin
  Result := TTextHintOverlay.Create(nil, Self);
  Result.Caption := ACaption;
end;

procedure TTHEdit.SetTextHint(AValue: String);
begin
  if (FTextHint = AValue) then Exit;
  FTextHint := AValue;
  TextHintOverlay.Caption := FTextHint;
  if (Text = '') and (FTextHint <> '') then ShowTextHint;
end;

procedure TTHEdit.ShowTextHint;
begin
  if not (csDestroying in ComponentState) and
   Assigned(TextHintOverlay) then TextHintOverlay.ShowOverlay;
end;

procedure TTHEdit.HideTextHint;
begin
  if Assigned(TextHintOverlay) then TextHintOverlay.HideOverlay;
end;

function TTHEdit.TextHintVisible: Boolean;
begin
  Result := TextHintOverlay.Visible;
end;

procedure TTHEdit.DoEnter;
begin
  inherited DoEnter;
  HideTextHint;
end;

procedure TTHEdit.DoExit;
begin
  inherited DoExit;
  if (Text = '') then ShowTextHint;
end;

procedure TTHEdit.DoSetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited DoSetBounds(ALeft, ATop, AWidth, AHeight);
  if TextHintVisible then
    ShowTextHint;   //repositions the overlay
end;

procedure TTHEdit.Loaded;
begin
  inherited Loaded;
  TextHintOverlay.Caption := FTextHint;
  if (Text = '') then TextHintOverlay.ShowOverlay;
end;

procedure TTHEdit.TextChanged;
begin
  inherited TextChanged;
  if (Text = '') and (FTextHint <> '') and not Focused then ShowTextHint;
end;

procedure TTHEdit.SetParent(NewParent: TWinControl);
var
  OldParent: TWinControl;
begin
  //writeln('TTHEdit.SetParent A');
  OldParent := Parent;
  //GTK crashes if TextHintVisible ...
  if not (csDestroying in ComponentState) then
    FreeAndNil(TextHintOverLay);
  //writeln('TTHEdit.SetParent B');
  inherited SetParent(NewParent);
  //writeln('TTHEdit.SetParent C');
  if not (csDestroying in ComponentState) then
  begin
    TextHintOverlay := CreateTextHintOverlay(FTextHint);
    if (OldParent <> Parent) and Assigned(Parent) and (Text = '') then ShowTextHint;
  end;
  //writeln('TTHEdit.SetParent End.');
end;



constructor TTHEdit.Create(AOwner: TComponent);
begin
  TextHintOverlay := CreateTextHintOverlay('');
  inherited Create(AOwner);
end;

destructor TTHEdit.Destroy;
begin
  if assigned(TextHintOverlay) then TextHintOverlay.Free;
  inherited Destroy;
end;



procedure TForm1.Button1Click(Sender: TObject);
var
  i: Integer;
  C: TControl;
begin
  for i := 0 to ControlCount - 1 do
  begin
    C := Controls[i];
    //writeln(Format('Controls[%d] = %s',[i,DbgSName(C)]));
    if C is TTextHintOverlay then
    begin
      C.SendToBack;
      //writeln(Format('   Controls[%d] is TTextHintOverlay',[i]));
    end;
  end;
  writeln;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  THEdit.TextHint := Edit1.Text;//THEdit.TextHint + '-##'
end;

procedure TForm1.Button3Click(Sender: TObject);
begin
  if THEdit.Parent = Self then
  begin
    THEdit.Parent := Panel1;
    THEdit.Left := 5;
    THEdit.Top := 5;
  end
  else
  begin
    THEdit.Parent := Self;
    THEdit.Top := Memo1.Top;
    THEdit.Left := Memo1.Left + memo1.Width + 10;
  end;
end;


procedure TForm1.Button4Click(Sender: TObject);
begin
  THEdit.Text := '';
end;

procedure TForm1.Button5Click(Sender: TObject);
var
  cl: TColor;
  Idx: Integer;
  RGBValue, RGB2: LongInt;
  CRP: TColor;
  CRPRGB: TColor;
begin
  cl := Edit1.Color;
  Idx := SysColorToSysColorIndex(cl);
  RGBValue := ColorToRGB(cl);
  RGB2 := ColorToRGB(clWindowText);
  CRP := Edit1.GetColorResolvingParent;
  CRPRGB := Edit1.GetRGBColorResolvingParent;
  writeln('Edit1.Color  = ',IntTohex(cl,8));
  writeln('clWhite      = ',IntToHex(clWhite,8));
  writeln('clWindowText = ',IntToHex(clWindowText,8));
  writeln('clDefault    = ',IntToHex(clDefault,8));
  writeln('Idx          = ',Idx);
  writeln('RGBValue     = ',IntTohex(RGBValue,8));
  writeln('RGB2         = ',IntTohex(RGB2,8));
  writeln('CRP          = ',IntTohex(CRP,8));
  writeln('CRPRGB       = ',IntTohex(CRPRGB,8));
end;

procedure TForm1.Edit1Enter(Sender: TObject);
begin
  THO.HideOverlay;
end;

procedure TForm1.Edit1Exit(Sender: TObject);
begin
  if (Edit1.Text = '') then THO.ShowOverlay;
end;

procedure TForm1.Edit1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  writeln('TForm1.Edit1MouseDown');
end;

procedure TForm1.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  THEdit := TTHedit.Create(Self);
  THEdit.Top := Memo1.Top;
  THEdit.Left := Memo1.Left + memo1.Width + 10;
  THEdit.TextHint := 'TextHint';
  THEdit.Width := Edit1.Width;
  THEdit.Parent := Self;

  THO := TTextHintOverlay.Create(nil, TCustomEdit(Edit1));
  THO.Top := THEdit.Top + THEdit.Height + 10;
  THO.Left := THEdit.Left;
  THO.Caption := '&THO_Caption'+LineEnding+'Line2';
  //THO.Parent := Self;

  //writeln('Edit1.BorderWidth = ',Edit1.BorderWidth);
  THEdit.SendToBack;

end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  THO.Free;
end;

initialization
end.

{

C:\Users\Bart\LazarusProjecten>test
[TCustomMemo.InitializeWnd] A: Lines:TTextStrings
[TCustomMemo.InitializeWnd] END Memo1:TMemo Lines:TWin32MemoStrings Lines.Count=
1
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"  (Last)
=====================
TWin32WSCustomMemo.SetWordWrap Before RecreateWnd: NewWordWrap=True
[TCustomMemo.FinalizeWnd] A Memo1:TMemo Lines:TWin32MemoStrings Lines.Count=1
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB" (Last)
[TCustomMemo.FinalizeWnd] END Memo1:TMemo Lines:TTextStrings Lines.Count=1
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB" (Last)
[TCustomMemo.InitializeWnd] A: Lines:TTextStrings
[TCustomMemo.InitializeWnd] END Memo1:TMemo Lines:TWin32MemoStrings Lines.Count=
2
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAA"
  Lines[1] = "BBBBBBBBBBBBBBBBBBBBBBBBBBBBB"  (Last)
TWin32WSCustomMemo.SetWordWrap After RecreateWnd: NewWordWrap=True
=====================
=====================
TWin32WSCustomMemo.SetWordWrap Before RecreateWnd: NewWordWrap=False
[TCustomMemo.FinalizeWnd] A Memo1:TMemo Lines:TWin32MemoStrings Lines.Count=2
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAA"
  Lines[1] = "BBBBBBBBBBBBBBBBBBBBBBBBBBBBB" (Last)
[TCustomMemo.FinalizeWnd] END Memo1:TMemo Lines:TTextStrings Lines.Count=2
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
  Lines[1] = "" (Last)
[TCustomMemo.InitializeWnd] A: Lines:TTextStrings
[TCustomMemo.InitializeWnd] END Memo1:TMemo Lines:TWin32MemoStrings Lines.Count=
2
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
  Lines[1] = ""  (Last)
TWin32WSCustomMemo.SetWordWrap After RecreateWnd: NewWordWrap=False
=====================
[TCustomMemo.FinalizeWnd] A Memo1:TMemo Lines:TWin32MemoStrings Lines.Count=1
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB" (Last)
[TCustomMemo.FinalizeWnd] END Memo1:TMemo Lines:TTextStrings Lines.Count=1
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
  Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB" (Last)







  =====================
  TWin32WSCustomMemo.SetWordWrap Before RecreateWnd: NewWordWrap=False
  [TCustomMemo.FinalizeWnd] A Memo1:TMemo Lines:TWin32MemoStrings Lines.Count=2
    Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAA"
    Lines[1] = "BBBBBBBBBBBBBBBBBBBBBBBBBBBBB" (Last)
  TTextStrings.AddStrings A: TheStrings.Count = 2
  TTextStrings.AddStrings End: Self.Count = 2
  [TCustomMemo.FinalizeWnd] B Memo1:TMemo NewStrings:TTextStrings NewStrings.Count
  =2
    NewStrings[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
    NewStrings[1] = "" (Last)
  [TCustomMemo.FinalizeWnd] D Memo1:TMemo Lines:TTextStrings Lines.Count=2
    Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
    Lines[1] = "" (Last)
  [TCustomMemo.FinalizeWnd] END Memo1:TMemo Lines:TTextStrings Lines.Count=2
    Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
    Lines[1] = "" (Last)
  [TCustomMemo.InitializeWnd] A: Lines:TTextStrings
  [TCustomMemo.InitializeWnd] END Memo1:TMemo Lines:TWin32MemoStrings Lines.Count=
  2
    Lines[0] = "AAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
    Lines[1] = ""  (Last)
  TWin32WSCustomMemo.SetWordWrap After RecreateWnd: NewWordWrap=False
  =====================
}
